==> abacusai-1.2.4.tar.gz: abacusai-1.2.4/abacusai/api_class/abstract.py <==
import dataclasses
import datetime
import inspect
import re
from abc import ABC
from copy import deepcopy
from textwrap import dedent
from typing import Any, Callable, get_origin, get_type_hints

from .enums import ApiEnum


FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')


def camel_case(value):
    if value:
        components = value.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])
    return value


def upper_snake_case(value):
    if value:
        s1 = FIRST_CAP_RE.sub(r'\1_\2', value)
        return ALL_CAP_RE.sub(r'\1_\2', s1).upper()
    return value


def snake_case(value):
    if value:
        s1 = FIRST_CAP_RE.sub(r'\1_\2', value)
        return ALL_CAP_RE.sub(r'\1_\2', s1).lower()
    return value


def get_clean_function_source_code(func: Callable):
    sample_lambda = (lambda: 0)
    if isinstance(func, type(sample_lambda)) and func.__name__ == sample_lambda.__name__:
        raise ValueError('Lambda function not allowed.')
    source_code = inspect.getsource(func)
    # If function source code has some initial indentation, remove it (Ex - can happen if the functor was defined inside a function)
    source_code = dedent(source_code)
    return source_code


@dataclasses.dataclass
class ApiClass(ABC):
    _upper_snake_case_keys: bool = dataclasses.field(default=False, repr=False, init=False)
    _support_kwargs: bool = dataclasses.field(default=False, repr=False, init=False)

    def __post_init__(self):
        if inspect.isabstract(self):
            raise ValueError(f'Cannot instantiate abstract class {self.__class__.__name__}')

    @classmethod
    def _get_builder(cls):
        return None

    def __str__(self):
        non_none_fields = {key: value for key, value in self.__dict__.items() if value is not None}
        field_str = ', '.join([f'{key}={value!r}' for key, value in non_none_fields.items()])
        return f'{self.__class__.__name__}({field_str})'

    def _repr_html_(self):
        return self.__str__()

    def __getitem__(self, item: str):
        if hasattr(self, item):
            return getattr(self, item)
        elif hasattr(self, snake_case(item)):
            return getattr(self, snake_case(item))
        elif self._support_kwargs and hasattr(self, 'kwargs') and self.kwargs is not None and (self.kwargs.get(item) is not None):
            return self.kwargs.get(item)
        else:
            raise KeyError(f'Key: {item} is not a property of {self.__class__.__name__}')

    def __setitem__(self, item: str, value: Any):
        if hasattr(self, item):
            setattr(self, item, value)
        elif hasattr(self, snake_case(item)):
            setattr(self, snake_case(item), value)
        elif self._support_kwargs:
            self.kwargs[item] = value
        else:
            raise KeyError(f'Key: {item} is not a property of {self.__class__.__name__}')

    def _unset_item(self, item: str):
        if hasattr(self, item):
            setattr(self, item, None)
        elif hasattr(self, snake_case(item)):
            setattr(self, snake_case(item), None)
        elif self._support_kwargs and hasattr(self, 'kwargs') and self.kwargs is not None and (self.kwargs.get(item) is not None):
            self.kwargs.pop(item)
        else:
            raise KeyError(f'Key: {item} is not a property of {self.__class__.__name__}')

    def get(self, item: str, default: Any = None):
        try:
            return self.__getitem__(item)
        except KeyError:
            return default

    def pop(self, item: str, default: Any = NotImplemented):
        try:
            value = self.__getitem__(item)
            self._unset_item(item)
            return value
        except KeyError:
            if default is NotImplemented:
                raise
            return default

    def to_dict(self):
        """
        Standardizes converting an ApiClass to dictionary.
        Keys of response dictionary are converted to camel case.
        This also validates the fields ( type, value, etc ) received in the dictionary.
        """
        def to_dict_helper(api_class_obj):
            res = {}
            api_class_dict = vars(api_class_obj)
            if self._support_kwargs:
                kwargs = api_class_dict.get('kwargs', None)
                api_class_dict.update(kwargs or {})
            for k, v in api_class_dict.items():
                if v is not None and k != 'kwargs':
                    if not k.startswith('__'):
                        k = upper_snake_case(k) if self._upper_snake_case_keys else camel_case(k)
                    if isinstance(v, ApiClass):
                        res[k] = to_dict_helper(v)
                    elif isinstance(v, list):
                        res[k] = [to_dict_helper(val) if isinstance(val, ApiClass) else val for val in v]
                    elif isinstance(v, dict):
                        res[k] = {key: to_dict_helper(val) if isinstance(val, ApiClass) else val for key, val in v.items()}
                    elif isinstance(v, datetime.datetime) or isinstance(v, datetime.date):
                        res[k] = v.isoformat() if v else v
                    elif isinstance(v, ApiEnum):
                        res[k] = v.value
                    else:
                        res[k] = v
            return res

        return to_dict_helper(self)

    @classmethod
    def from_dict(cls, input_dict: dict):
        if input_dict is None:
            return None
        obj = None
        if input_dict:
            if builder := cls._get_builder():
                config_class_key = None
                value = next((key for key, val in builder.config_class_map.items() if val.__name__ == cls.__name__), None)
                input_dict_with_config_key = input_dict
                if value is not None:
                    input_dict_with_config_key = deepcopy(input_dict)
                    if builder.config_abstract_class._upper_snake_case_keys:
                        config_class_key = upper_snake_case(builder.config_class_key)
                        if config_class_key not in input_dict_with_config_key:
                            input_dict_with_config_key[config_class_key] = value
                    else:
                        config_class_key = builder.config_class_key
                        if config_class_key not in input_dict_with_config_key and camel_case(config_class_key) not in input_dict_with_config_key:
                            input_dict_with_config_key[config_class_key] = value

                obj = builder.from_dict(input_dict_with_config_key)

            if not cls._upper_snake_case_keys:
                input_dict = {snake_case(k): v for k, v in input_dict.items()}
            if not cls._support_kwargs:
                # only use keys that are valid fields in the ApiClass
                field_names = set((field.name) for field in dataclasses.fields(cls))
                input_dict = {k: v for k, v in input_dict.items() if k in field_names}
        if obj is None:
            obj = cls(**input_dict)

        for attr_name, attr_type in get_type_hints(cls).items():
            if attr_name in input_dict and inspect.isclass(attr_type) and issubclass(attr_type, ApiClass):
                setattr(obj, attr_name, attr_type.from_dict(input_dict[attr_name]))
            elif attr_name in input_dict and get_origin(attr_type) is list and attr_type.__args__ and inspect.isclass(attr_type.__args__[0]) and issubclass(attr_type.__args__[0], ApiClass):
                class_type = attr_type.__args__[0]
                if isinstance(input_dict[attr_name], list):
                    setattr(obj, attr_name, [class_type.from_dict(item) for item in input_dict[attr_name]])
                else:
                    raise ValueError(f'Expected list for {attr_name} but got {type(input_dict[attr_name])}')

        return obj


@dataclasses.dataclass
class _ApiClassFactory(ABC):
    config_abstract_class = None
    config_class_key = None
    config_class_map = {}

    @classmethod
    def from_dict(cls, config: dict) -> ApiClass:
        support_kwargs = cls.config_abstract_class and cls.config_abstract_class._support_kwargs
        is_upper_snake_case_keys = cls.config_abstract_class and cls.config_abstract_class._upper_snake_case_keys
        config_class_key = upper_snake_case(cls.config_class_key) if is_upper_snake_case_keys else cls.config_class_key
        # Logic here is that the we keep the config_class_key in snake_case if _upper_snake_case_keys is False else we convert it to upper_snake_case
        # if _upper_snake_case_keys is False then we check in both casing: 1. snake_case and 2. camel_case
        if not is_upper_snake_case_keys and config_class_key not in config and camel_case(config_class_key) in config:
            config_class_key = camel_case(config_class_key)

        if not support_kwargs and config_class_key not in (config or {}):
            raise KeyError(f'Could not find {config_class_key} in {config}')
        config_class_type = config.get(config_class_key, None)

        if isinstance(config_class_type, str):
            config_class_type = config_class_type.upper()
        config_class = cls.config_class_map.get(config_class_type)

        if support_kwargs:
            if config_class:
                field_names = set((field.name) for field in dataclasses.fields(config_class))
                trimmed_config = {}
                kwargs = {}
                for k, v in config.items():
                    if snake_case(k) in field_names:
                        trimmed_config[snake_case(k)] = v
                    else:
                        kwargs[k] = v
                if len(kwargs):
                    trimmed_config['kwargs'] = kwargs
                config = trimmed_config
            else:
                config = {'kwargs': config}
                config_class = cls.config_abstract_class
        if config_class is None:
            raise ValueError(f'Invalid type {config_class_type}')
        if not config_class._upper_snake_case_keys:
            config = {snake_case(k): v for k, v in config.items()}
        config.pop(snake_case(config_class_key), None)
        return config_class(**config)

==> abacusai-1.3.0.tar.gz: abacusai-1.3.0/abacusai/api_class/abstract.py <==
import dataclasses
import datetime
import inspect
import re
from abc import ABC
from copy import deepcopy
from textwrap import dedent
from typing import Any, Callable, get_origin, get_type_hints

from .enums import ApiEnum


FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')


def camel_case(value):
    if value:
        components = value.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])
    return value


def upper_snake_case(value):
    if value:
        s1 = FIRST_CAP_RE.sub(r'\1_\2', value)
        return ALL_CAP_RE.sub(r'\1_\2', s1).upper()
    return value


def snake_case(value):
    if value:
        s1 = FIRST_CAP_RE.sub(r'\1_\2', value)
        return ALL_CAP_RE.sub(r'\1_\2', s1).lower()
    return value


def get_clean_function_source_code(func: Callable):
    sample_lambda = (lambda: 0)
    if isinstance(func, type(sample_lambda)) and func.__name__ == sample_lambda.__name__:
        raise ValueError('Lambda function not allowed.')
    source_code = inspect.getsource(func)
    # If function source code has some initial indentation, remove it (Ex - can happen if the functor was defined inside a function)
    source_code = dedent(source_code)
    return source_code


@dataclasses.dataclass
class ApiClass(ABC):
    _upper_snake_case_keys: bool = dataclasses.field(default=False, repr=False, init=False)
    _support_kwargs: bool = dataclasses.field(default=False, repr=False, init=False)

    def __post_init__(self):
        if inspect.isabstract(self):
            raise ValueError(f'Cannot instantiate abstract class {self.__class__.__name__}')

    @classmethod
    def _get_builder(cls):
        return None

    def __str__(self):
        non_none_fields = {key: value for key, value in self.__dict__.items() if value is not None}
        field_str = ', '.join([f'{key}={value!r}' for key, value in non_none_fields.items()])
        return f'{self.__class__.__name__}({field_str})'

    def _repr_html_(self):
        return self.__str__()

    def __getitem__(self, item: str):
        if hasattr(self, item):
            return getattr(self, item)
        elif hasattr(self, snake_case(item)):
            return getattr(self, snake_case(item))
        elif self._support_kwargs and hasattr(self, 'kwargs') and self.kwargs is not None and (self.kwargs.get(item) is not None):
            return self.kwargs.get(item)
        else:
            raise KeyError(f'Key: {item} is not a property of {self.__class__.__name__}')

    def __setitem__(self, item: str, value: Any):
        if hasattr(self, item):
            setattr(self, item, value)
        elif hasattr(self, snake_case(item)):
            setattr(self, snake_case(item), value)
        elif self._support_kwargs:
            self.kwargs[item] = value
        else:
            raise KeyError(f'Key: {item} is not a property of {self.__class__.__name__}')

    def _unset_item(self, item: str):
        if hasattr(self, item):
            setattr(self, item, None)
        elif hasattr(self, snake_case(item)):
            setattr(self, snake_case(item), None)
        elif self._support_kwargs and hasattr(self, 'kwargs') and self.kwargs is not None and (self.kwargs.get(item) is not None):
            self.kwargs.pop(item)
        else:
            raise KeyError(f'Key: {item} is not a property of {self.__class__.__name__}')

    def get(self, item: str, default: Any = None):
        try:
            return self.__getitem__(item)
        except KeyError:
            return default

    def pop(self, item: str, default: Any = NotImplemented):
        try:
            value = self.__getitem__(item)
            self._unset_item(item)
            return value
        except KeyError:
            if default is NotImplemented:
                raise
            return default

    def to_dict(self):
        """
        Standardizes converting an ApiClass to dictionary.
        Keys of response dictionary are converted to camel case.
        This also validates the fields ( type, value, etc ) received in the dictionary.
        """
        def to_dict_helper(api_class_obj):
            res = {}
            api_class_dict = vars(api_class_obj)
            if self._support_kwargs:
                kwargs = api_class_dict.get('kwargs', None)
                api_class_dict.update(kwargs or {})
            for k, v in api_class_dict.items():
                if v is not None and k != 'kwargs':
                    if not k.startswith('__'):
                        k = upper_snake_case(k) if self._upper_snake_case_keys else camel_case(k)
                    if isinstance(v, ApiClass):
                        res[k] = to_dict_helper(v)
                    elif isinstance(v, list):
                        res[k] = [to_dict_helper(val) if isinstance(val, ApiClass) else val for val in v]
                    elif isinstance(v, dict):
                        res[k] = {key: to_dict_helper(val) if isinstance(val, ApiClass) else val for key, val in v.items()}
                    elif isinstance(v, datetime.datetime) or isinstance(v, datetime.date):
                        res[k] = v.isoformat() if v else v
                    elif isinstance(v, ApiEnum):
                        res[k] = v.value
                    else:
                        res[k] = v
            return res

        return to_dict_helper(self)

    @classmethod
    def from_dict(cls, input_dict: dict):
        if input_dict is None:
            return None
        obj = None
        if input_dict:
            builder = cls._get_builder()
            if builder:
                config_class_key = None
                value = next((key for key, val in builder.config_class_map.items() if val.__name__ == cls.__name__), None)
                input_dict_with_config_key = input_dict
                if value is not None:
                    input_dict_with_config_key = deepcopy(input_dict)
                    if builder.config_abstract_class._upper_snake_case_keys:
                        config_class_key = upper_snake_case(builder.config_class_key)
                        if config_class_key not in input_dict_with_config_key:
                            input_dict_with_config_key[config_class_key] = value
                    else:
                        config_class_key = builder.config_class_key
                        if config_class_key not in input_dict_with_config_key and camel_case(config_class_key) not in input_dict_with_config_key:
                            input_dict_with_config_key[config_class_key] = value

                obj = builder.from_dict(input_dict_with_config_key)

            if not cls._upper_snake_case_keys:
                input_dict = {snake_case(k): v for k, v in input_dict.items()}
            if not cls._support_kwargs:
                # only use keys that are valid fields in the ApiClass
                field_names = set((field.name) for field in dataclasses.fields(type(obj) if obj else cls))
                input_dict = {k: v for k, v in input_dict.items() if k in field_names}
        if obj is None:
            obj = cls(**input_dict)

        for attr_name, attr_type in get_type_hints(type(obj)).items():
            if attr_name in input_dict and isinstance(input_dict[attr_name], dict) and inspect.isclass(attr_type) and issubclass(attr_type, ApiClass):
                setattr(obj, attr_name, attr_type.from_dict(input_dict[attr_name]))
            elif attr_name in input_dict and get_origin(attr_type) is list and attr_type.__args__ and inspect.isclass(attr_type.__args__[0]) and issubclass(attr_type.__args__[0], ApiClass):
                class_type = attr_type.__args__[0]
                if isinstance(input_dict[attr_name], list):
                    if not all(isinstance(item, dict) or isinstance(item, class_type) for item in input_dict[attr_name]):
                        raise ValueError(attr_name, f'Expected list of {class_type} or dictionary for {attr_name}')
                    setattr(obj, attr_name, [item if isinstance(item, class_type) else class_type.from_dict(item) for item in input_dict[attr_name]])
                else:
                    raise ValueError(attr_name, f'Expected list for {attr_name} but got {type(input_dict[attr_name])}')

        return obj


@dataclasses.dataclass
class _ApiClassFactory(ABC):
    config_abstract_class = None
    config_class_key = None
    config_class_map = {}

    @classmethod
    def from_dict(cls, config: dict) -> ApiClass:
        support_kwargs = cls.config_abstract_class and cls.config_abstract_class._support_kwargs
        is_upper_snake_case_keys = cls.config_abstract_class and cls.config_abstract_class._upper_snake_case_keys
        config_class_key = upper_snake_case(cls.config_class_key) if is_upper_snake_case_keys else cls.config_class_key
        # Logic here is that the we keep the config_class_key in snake_case if _upper_snake_case_keys False else we convert it to upper_snake_case
        # if _upper_snake_case_keys is False then we check in both casing: 1. snake_case and 2. camel_case
        if not is_upper_snake_case_keys and config_class_key not in config and camel_case(config_class_key) in config:
            config_class_key = camel_case(config_class_key)

        if not support_kwargs and config_class_key not in (config or {}):
            raise KeyError(f'Could not find {config_class_key} in {config}')
        config_class_type = config.get(config_class_key, None)

        if isinstance(config_class_type, str):
            config_class_type = config_class_type.upper()
        config_class = cls.config_class_map.get(config_class_type)

        if support_kwargs:
            if config_class:
                field_names = set((field.name) for field in dataclasses.fields(config_class))
                trimmed_config = {}
                kwargs = {}
                for k, v in config.items():
                    if snake_case(k) in field_names:
                        trimmed_config[snake_case(k)] = v
                    else:
                        kwargs[k] = v
                if len(kwargs):
                    trimmed_config['kwargs'] = kwargs
                config = trimmed_config
            else:
                config = {'kwargs': config}
                config_class = cls.config_abstract_class
        if config_class is None:
            raise ValueError(f'Invalid type {config_class_type}')
        if not config_class._upper_snake_case_keys:
            config = {snake_case(k): v for k, v in config.items()}
        config.pop(snake_case(config_class_key), None)
        return config_class(**config)
